/*
 * BubsyMOD source code - audio_mixer.c
 * This application is open source and can be freely distributed
 * and modified, just leave credits to 8bitbubsy and mukunda.
 *
 * audio_mixer.c written by mukunda and 8bitbubsy
 * bubsymod.c written by 8bitbubsy
 * pattern_viewer.c written by 8bitbubsy
 * protracker_loader.c written by 8bitbubsy
 * protracker_sequencer.c written by mukunda and 8bitbubsy
 * sample_viewer.c written by 8bitbubsy
 * textout.c written by 8bitbubsy
 * volume.c written by 8bitbubsy
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL_audio.h>
#include "bubsymod.h"

/* fractional precision, code taken from ScummVM (/common/frac.h) */
#define FRAC_BITS 14
#define intToFrac(x) (unsigned int)((x) << FRAC_BITS)
#define fracToInt(x) (unsigned int)((x) >> FRAC_BITS)
#define doubleToFrac(x) (unsigned int)((x) * (1L << FRAC_BITS))
/* -------------------------------------------------------------- */

static int mixbufL[SOUND_BUFFER_SIZE], mixbufR[SOUND_BUFFER_SIZE];

typedef struct
{
	const char *data;      /* voice data     - -128 to 127   -  8 bits */
	unsigned char vol;      /* voice volume   -  0   to 64    -  7 bits */
	unsigned int panL;     /* voice panning  -  0   to 255   -  8 bits */
	unsigned int panR;     /* voice panning  -  0   to 255   -  8 bits */
	unsigned int rate;     /* voice inc rate -  0   to 65535 - 16 bits */
	unsigned int len;      /* voice read len -  0   to 65535 - 16 bits */
	unsigned int loop_len; /* voice loop len -  0   to 65535 - 16 bits */
	unsigned int loop_end; /* voice loop end -  0   to 65535 - 16 bits */

	unsigned int   pos;      /* voice position */
} Voice;

Voice v[4];

int bubsymodPulse(int samples); /* declared in protracker_sequencer.c */

void audioChangeSrc(int ch, const char *src, unsigned int len, unsigned int loop_pos, unsigned int loop_len)
{
	if (loop_len > 2) /* <= 2 is the same as no loop, Amiga is weird */
	{
		if (v[ch].pos >= (loop_pos + loop_len))
			v[ch].pos = loop_pos + loop_len;

		v[ch].loop_end = loop_pos + loop_len;
		v[ch].loop_len = loop_len;
	}
	else
	{
		if (v[ch].pos >= len)
			v[ch].pos = 0;

		v[ch].loop_end = 0;
		v[ch].loop_len = 0;
	}

	v[ch].data = src;
	v[ch].len = len;

	return;
}

void audioSetSource(int ch, const char *src, unsigned int len, unsigned int loop_pos, unsigned int loop_len)
{
	v[ch].data = src;
	v[ch].len = len;
	v[ch].pos = 0; /* always reset the voice position on set source */
	v[ch].loop_end = loop_pos + loop_len;
	v[ch].loop_len = loop_len;
	
	return;
}

void audioSetPanning(int ch, unsigned char pan)
{
	v[ch].panL = 255 - pan;
	v[ch].panR = pan;

	return;
}

void audioCutAllVoices()
{
	memset(v, 0, sizeof (Voice) * 4);

	/* amiga paula panning, but softer */
	audioSetPanning(0, 64);
	audioSetPanning(1, 192);
	audioSetPanning(2, 192);
	audioSetPanning(3, 64);
	
	return;
}

void audioSetOffset(int ch, unsigned int pos)
{
	if (pos < v[ch].len)
		v[ch].pos = intToFrac(pos);
		
	return;
}

void audioSetVolume(int ch, int vol)
{
	v[ch].vol = vol;
	return;
}

void audioSetRate(int ch, int rate)
{
	v[ch].rate = rate;
	return;
}

void audioPause()
{
	audioSetSource(0, NULL, 0, 0, 0);
	audioSetSource(1, NULL, 0, 0, 0);
	audioSetSource(2, NULL, 0, 0, 0);
	audioSetSource(3, NULL, 0, 0, 0);
	
	return;
}

static void OutputAudio(signed short *target, int nsamples)
{
	signed short *out;

	/* clear sample buffer */
	memset(mixbufL, 0, nsamples * sizeof (int));
	memset(mixbufR, 0, nsamples * sizeof (int));

	/* mix the samples, amplify by volume and panning, check loop points */
	{
		int ch;
		for (ch = 0; ch < 4; ch++)
		{
			if (v[ch].data != NULL)
			{
				int smp;
				for (smp = 0; smp < nsamples; smp++)
				{
					/* the sample data is 8 bits, the volume is 7 bits */
					int t_smp = v[ch].data[fracToInt(v[ch].pos)] * v[ch].vol;

					/* t_smp is 15 bits, the panning is 8 bits */
					mixbufL[smp] += (t_smp * v[ch].panL); /* 23 bits */
					mixbufR[smp] += (t_smp * v[ch].panR); /* 23 bits */

					v[ch].pos += v[ch].rate;

					if (fracToInt(v[ch].pos) >= v[ch].loop_end)
						v[ch].pos -= intToFrac(v[ch].loop_len);

					if (fracToInt(v[ch].pos) >= v[ch].len)
					{
						v[ch].data = NULL;

						break;
					}
				}
			}
		}
	}

	out = target;

	/* shift from 23 bits to 16 bits, to prevent ear damage! */
	{
		int s;
		for (s = 0; s < nsamples; s++)
		{
			int inL = mixbufL[s] >> 7;
			int inR = mixbufR[s] >> 7;

			/* saturate */
			if (inL < -32768) inL = -32768;
			else if (inL > 32767) inL = 32767;

			if (inR < -32768) inR = -32768;
			else if (inR > 32767) inR = 32767;

			*out++ = (signed short)inL;
			*out++ = (signed short)inR;
		}
	}
	
	return;
}

void mixaudio(void *userdata, unsigned char *stream, int len)
{
	/* SDL mix audio callback routine */
	signed short *out = (signed short *)stream;

	int samples = len / 4;
	while (samples)
	{
		/* find out how many samples we need to mix */
		int gen = bubsymodPulse(samples);
		OutputAudio(out, gen);

		out += (gen * 2);
		samples -= gen;
	}
	
	return;
}

static int OpenDev()
{
	SDL_AudioSpec fmt;

	/* 16 bit stereo */
	fmt.freq = SOUND_FREQUENCY;
	fmt.format = AUDIO_S16SYS;
	fmt.channels = 2;
	fmt.samples = SOUND_BUFFER_SIZE;
	fmt.callback = mixaudio;
	fmt.userdata = NULL;

	if (SDL_OpenAudio(&fmt, NULL) < 0)
	{
		fprintf(stderr, "Unable to open audio: %s\n", SDL_GetError());

		return 0;
	}

	return 1;
}

static void CloseDev()
{
	SDL_CloseAudio();
	return;
}

int audioStart() 
{
	audioCutAllVoices();

	/* initialize sound mixer */
	if (!OpenDev())
		return 0;

	/* unpause audio */
	SDL_PauseAudio(0);

	return 1;
}

void audioStop()
{
	audioCutAllVoices();
	
	SDL_Delay(100);

	/* pause before closing (sanity) */
	SDL_PauseAudio(1);

	/* close mixer */
	CloseDev();

	return;
}
