/*
 * BubsyMOD source code - protracker_loader.c
 * This application is open source and can be freely distributed
 * and modified, just leave credits to 8bitbubsy and mukunda.
 *
 * audio_mixer.c written by mukunda and 8bitbubsy
 * bubsymod.c written by 8bitbubsy
 * pattern_viewer.c written by 8bitbubsy
 * protracker_loader.c written by 8bitbubsy
 * protracker_sequencer.c written by mukunda and 8bitbubsy
 * sample_viewer.c written by 8bitbubsy
 * textout.c written by 8bitbubsy
 * volume.c written by 8bitbubsy
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bubsymod.h"

static int get_word_bigendian(FILE *fmodule)
{
	unsigned char b0;
	unsigned char b1;

	fread(&b0, 1, 1, fmodule);
	fread(&b1, 1, 1, fmodule);
	
	return (unsigned short)((b0 << 8) + b1);
}

static int clamp(int min, int max, int val)
{
	if (val < min)
		return min;
	if (val > max)
		return max;
		
	return val;
}

extern char *file_sample_data;

static int check_mod_tag(const char *buf);

MODULE *bubsymodLoad(const char *filename)
{
	FILE *fmodule = NULL;
	MODULE *new_module = NULL;
	
	/* allocate space for module struct and clear it */
	new_module = (MODULE *)calloc(1, sizeof (MODULE));
	if (new_module == NULL) return NULL; /* out of memory */

	/* load our module (if it is any.. heh) */
	fmodule = fopen(filename, "rb");
	if (fmodule == NULL)
	{
		free(new_module);

		return NULL; /* can't be opened */
	}
	
	/* check module tag */
	{
		char modsig[5];

		fseek(fmodule, 0x0438, SEEK_SET);
		fread(modsig, 1, 4, fmodule);
		if (check_mod_tag(modsig) == 0)
		{
			free(new_module);

			return NULL; /* sorry, not a proper .MOD module! */
		}
		rewind(fmodule);
	}

	/* read song name */
	fread(new_module->head.module_title, 1, 20, fmodule);
	
	/* read sample information */
	{
		int i;
		for (i = 0; i < 31; i++)
		{
			/* sample text */
			fread(new_module->samples[i].text, 1, 22, fmodule);
			
			/* read sample length (amiga byte order) */
			new_module->samples[i].length = get_word_bigendian(fmodule) << 1;

			/* read finetune */
			fread(&new_module->samples[i].finetune, 1, 1, fmodule);
			new_module->samples[i].finetune &= 0x0F;

			/* read volume and clamp it */
			fread(&new_module->samples[i].volume, 1, 1, fmodule);
			new_module->samples[i].volume = clamp(0, 64, new_module->samples[i].volume);

			/* read loop start (amiga byte order) */
			new_module->samples[i].loop_start = get_word_bigendian(fmodule) << 1;

			/* read loop length (amiga byte order) */
			new_module->samples[i].loop_length = get_word_bigendian(fmodule) << 1;

			/* loop length fix, stupid amigas! */
			if (new_module->samples[i].loop_length <= 2)
			{
				new_module->samples[i].loop_length = 0;
				new_module->samples[i].loop_start = 0;
			}

			new_module->head.total_sample_size += new_module->samples[i].length;
		}
	}

	/* read number of orders */
	fread(&new_module->head.order_count, 1, 1, fmodule);
	new_module->head.order_count = clamp(0, 128, new_module->head.order_count);

	/* read song restart position */
	fread(&new_module->head.restart_pos, 1, 1, fmodule);
	if (new_module->head.restart_pos >= 127)
	{
		new_module->play_loop = 0;
	}
	else
	{
		new_module->play_loop = 1;
		new_module->head.restart_pos = (unsigned char)clamp(0, new_module->head.order_count, new_module->head.restart_pos);
	}

	/* read sequences/orders */
	{
		int i;
		for (i = 0; i < 128; i++)
		{
			fread(&new_module->head.order[i], 1, 1, fmodule);
			if (new_module->head.order[i] > new_module->head.pattern_count)
				new_module->head.pattern_count = new_module->head.order[i];
		}
	}
	
	new_module->head.pattern_count++;

	/* skip file format tag (not used in 31 samples .MODs) */
	fseek(fmodule, 4, SEEK_CUR);

	/* load pattern data */
	{
		modnote_t *note = NULL;

		int pattern;
		for (pattern = 0; pattern < new_module->head.pattern_count; pattern++)
		{
			/* allocate data for each pattern */
			new_module->patterns[pattern] = note = (modnote_t *)calloc(64 * 4, sizeof (modnote_t));
			if (note == NULL)
			{
				fclose(fmodule);
				free(new_module);

				return NULL; /* out of memory */
			}

			{
				int row;
				for (row = 0; row < 64; row++)
				{
					int channel;
					for (channel = 0; channel < 4; channel++)
					{
						unsigned char b1, b2, b3, b4;

						/* read four bytes */
						fread(&b1, 1, 1, fmodule);
						fread(&b2, 1, 1, fmodule);
						fread(&b3, 1, 1, fmodule);
						fread(&b4, 1, 1, fmodule);

						/* unpack period and clamp it (Amiga Paula range) */
						{
							short temp_var = ((b1 & 0x0F) << 8) | b2;
							if (temp_var == 0)
								note->period  = 0; /* 0 = important, 0 = no note */
							else 
								note->period = (short)clamp(113, 856, temp_var);
						}

						/* unpack sample number */
						note->sample = (char)clamp(0, 31, (b1 & 0xF0) | (b3 >> 4));

						note->command = b3 & 0x0F;
						note->param = b4;

						/* pre-clamp effects, safety first! */
						switch (note->command)
						{
							/* 0x00 - 0xF0 */
							case 0x05: /* volslide + portamento */
							case 0x06: /* volslide + vibrato */
							case 0x0A: /* volume slide */
								note->param = (unsigned char)clamp(0x00, 0xF0, note->param);
							break;

							/* 0x00 - 0x?? */
							case 0x0B: /* position jump */
								note->param = (unsigned char)clamp(0, new_module->head.order_count, note->param);
							break;

							/* 0x00 - 0x40 */
							case 0x0C: /* set volume */
								note->param = (unsigned char)clamp(0x00, 0x40, note->param);
							break;

							default:
								break;
						}

						note++;
					}
				}
			}
		}
	}
	
	/* allocate memory for samples */
	file_sample_data = (char *)malloc(new_module->head.total_sample_size);
	if (file_sample_data == NULL)
	{
		fclose(fmodule);
		free(new_module);

		return NULL; /* out of memory */
	}

	/* load sample data */
	{
		unsigned int sample_offset = 0;

		int i;
		for (i = 0; i < 31; i++)
		{
			new_module->samples[i].offset = sample_offset;
			fread(&file_sample_data[sample_offset], new_module->samples[i].length, 1, fmodule);
			sample_offset += new_module->samples[i].length;
		}
	}

	/* make the main sample data pointer point to our allocated memory where the samples are */
	new_module->sample_data = file_sample_data;

	fclose(fmodule);

	new_module->module_loaded = 1;

	return new_module;
}

static int check_mod_tag(const char *buf)
{
	/* read MODFIL10.TXT for more information on this */
	if (strncmp(buf, "M.K.", 4) == 0)
		return 1;
	if (strncmp(buf, "M!K!", 4) == 0)
		return 1;
	if (strncmp(buf, "M&K!", 4) == 0)
		return 1;
	if (strncmp(buf, "N.T.", 4) == 0)
		return 1;
	if (strncmp(buf, "FLT4", 4) == 0)
		return 1;
	if (strncmp(buf, "4CHN", 4) == 0)
		return 1;

	return 0; /* unknown .MOD file */
}
